@@ -125,6 +125,7 @@ pub trait SerialEvents {
 /// Provides a no-op implementation of `SerialEvents` which can be used in situations that
 /// do not require logging or otherwise doing anything in response to the events defined
 /// as part of `SerialEvents`.
+#[derive(Debug)]
 pub struct NoEvents;
 
 impl SerialEvents for NoEvents {
@@ -152,6 +153,48 @@ impl<EV: SerialEvents> SerialEvents for Arc<EV> {
     }
 }
 
+/// The state of the Serial device.
+#[derive(Clone, Debug, PartialEq)]
+pub struct SerialState {
+    /// Divisor Latch Low Byte
+    pub baud_divisor_low: u8,
+    /// Divisor Latch High Byte
+    pub baud_divisor_high: u8,
+    /// Interrupt Enable Register
+    pub interrupt_enable: u8,
+    /// Interrupt Identification Register
+    pub interrupt_identification: u8,
+    /// Line Control Register
+    pub line_control: u8,
+    /// Line Status Register
+    pub line_status: u8,
+    /// Modem Control Register
+    pub modem_control: u8,
+    /// Modem Status Register
+    pub modem_status: u8,
+    /// Scratch Register
+    pub scratch: u8,
+    /// Transmitter Holding Buffer/Receiver Buffer
+    pub in_buffer: Vec<u8>,
+}
+
+impl Default for SerialState {
+    fn default() -> Self {
+        SerialState {
+            baud_divisor_low: DEFAULT_BAUD_DIVISOR_LOW,
+            baud_divisor_high: DEFAULT_BAUD_DIVISOR_HIGH,
+            interrupt_enable: DEFAULT_INTERRUPT_ENABLE,
+            interrupt_identification: DEFAULT_INTERRUPT_IDENTIFICATION,
+            line_control: DEFAULT_LINE_CONTROL,
+            line_status: DEFAULT_LINE_STATUS,
+            modem_control: DEFAULT_MODEM_CONTROL,
+            modem_status: DEFAULT_MODEM_STATUS,
+            scratch: DEFAULT_SCRATCH,
+            in_buffer: Vec::new(),
+        }
+    }
+}
+
 /// The serial console emulation is done by emulating a serial COM port.
 ///
 /// Each serial COM port (COM1-4) has an associated Port I/O address base and
@@ -212,6 +255,7 @@ impl<EV: SerialEvents> SerialEvents for Arc<EV> {
 ///     serial.enqueue_raw_bytes(input).unwrap();
 /// }
 /// ```
+#[derive(Debug)]
 pub struct Serial<T: Trigger, EV: SerialEvents, W: Write> {
     // Some UART registers.
     baud_divisor_low: u8,
@@ -268,7 +312,60 @@ impl<T: Trigger, W: Write> Serial<T, NoEvents, W> {
 }
 
 impl<T: Trigger, EV: SerialEvents, W: Write> Serial<T, EV, W> {
-    /// Creates a new `Serial` instance which writes the guest's output to
+    /// Creates a new `Serial` instance from a given `state`, which writes the guest's output to
+    /// `out`, uses `trigger` object to notify the driver about new
+    /// events, and invokes the `serial_evts` implementation of `SerialEvents`
+    /// during operation.
+    /// For creating the instance from a default state, [`with_events`](#method.with_events) method
+    /// can be used.
+    ///
+    /// # Arguments
+    /// * `state` - A reference to the state from which the `Serial` is constructed.
+    /// * `trigger` - The `Trigger` object that will be used to notify the driver
+    ///               about events.
+    /// * `serial_evts` - The `SerialEvents` implementation used to track the occurrence
+    ///                   of significant events in the serial operation logic.
+    /// * `out` - An object for writing guest's output to. In case the output
+    ///           is not of interest,
+    ///           [std::io::Sink](https://doc.rust-lang.org/std/io/struct.Sink.html)
+    ///           can be used here.
+    pub fn from_state(
+        state: &SerialState,
+        trigger: T,
+        serial_evts: EV,
+        out: W,
+    ) -> Result<Self, Error<T::E>> {
+        if state.in_buffer.len() > FIFO_SIZE {
+            return Err(Error::FullFifo);
+        }
+
+        let mut serial = Serial {
+            baud_divisor_low: state.baud_divisor_low,
+            baud_divisor_high: state.baud_divisor_high,
+            interrupt_enable: state.interrupt_enable,
+            interrupt_identification: state.interrupt_identification,
+            line_control: state.line_control,
+            line_status: state.line_status,
+            modem_control: state.modem_control,
+            modem_status: state.modem_status,
+            scratch: state.scratch,
+            in_buffer: VecDeque::from(state.in_buffer.clone()),
+            interrupt_evt: trigger,
+            events: serial_evts,
+            out,
+        };
+
+        if serial.is_thr_interrupt_enabled() && serial.is_thr_interrupt_set() {
+            serial.trigger_interrupt().map_err(Error::Trigger)?;
+        }
+        if serial.is_rda_interrupt_enabled() && serial.is_rda_interrupt_set() {
+            serial.trigger_interrupt().map_err(Error::Trigger)?;
+        }
+
+        Ok(serial)
+    }
+
+    /// Creates a new `Serial` instance from the default state, which writes the guest's output to
     /// `out`, uses `trigger` object to notify the driver about new
     /// events, and invokes the `serial_evts` implementation of `SerialEvents`
     /// during operation.
@@ -283,20 +380,24 @@ impl<T: Trigger, EV: SerialEvents, W: Write> Serial<T, EV, W> {
     ///           [std::io::Sink](https://doc.rust-lang.org/std/io/struct.Sink.html)
     ///           can be used here.
     pub fn with_events(trigger: T, serial_evts: EV, out: W) -> Self {
-        Serial {
-            baud_divisor_low: DEFAULT_BAUD_DIVISOR_LOW,
-            baud_divisor_high: DEFAULT_BAUD_DIVISOR_HIGH,
-            interrupt_enable: DEFAULT_INTERRUPT_ENABLE,
-            interrupt_identification: DEFAULT_INTERRUPT_IDENTIFICATION,
-            line_control: DEFAULT_LINE_CONTROL,
-            line_status: DEFAULT_LINE_STATUS,
-            modem_control: DEFAULT_MODEM_CONTROL,
-            modem_status: DEFAULT_MODEM_STATUS,
-            scratch: DEFAULT_SCRATCH,
-            in_buffer: VecDeque::new(),
-            interrupt_evt: trigger,
-            events: serial_evts,
-            out,
+        // Safe because we are using the default state that has an appropriately size input buffer
+        // and there are no pending interrupts to be triggered.
+        Self::from_state(&SerialState::default(), trigger, serial_evts, out).unwrap()
+    }
+
+    /// Returns the state of the Serial.
+    pub fn state(&self) -> SerialState {
+        SerialState {
+            baud_divisor_low: self.baud_divisor_low,
+            baud_divisor_high: self.baud_divisor_high,
+            interrupt_enable: self.interrupt_enable,
+            interrupt_identification: self.interrupt_identification,
+            line_control: self.line_control,
+            line_status: self.line_status,
+            modem_control: self.modem_control,
+            modem_status: self.modem_status,
+            scratch: self.scratch,
+            in_buffer: Vec::from(self.in_buffer.clone()),
         }
     }
 
@@ -322,6 +423,14 @@ impl<T: Trigger, EV: SerialEvents, W: Write> Serial<T, EV, W> {
         (self.interrupt_enable & IER_THR_EMPTY_BIT) != 0
     }
 
+    fn is_rda_interrupt_set(&self) -> bool {
+        (self.interrupt_identification & IIR_RDA_BIT) != 0
+    }
+
+    fn is_thr_interrupt_set(&self) -> bool {
+        (self.interrupt_identification & IIR_THR_EMPTY_BIT) != 0
+    }
+
     fn is_in_loop_mode(&self) -> bool {
         (self.modem_control & MCR_LOOP_BIT) != 0
     }
@@ -540,15 +649,19 @@ impl<T: Trigger, EV: SerialEvents, W: Write> Serial<T, EV, W> {
     pub fn enqueue_raw_bytes(&mut self, input: &[u8]) -> Result<usize, Error<T::E>> {
         let mut write_count = 0;
         if !self.is_in_loop_mode() {
+            // First check if the input slice and the fifo are non-empty so we can return early in
+            // those cases. Any subsequent `write` to the `in_buffer` will write at least one byte.
+            if input.is_empty() {
+                return Ok(0);
+            }
             if self.fifo_capacity() == 0 {
                 return Err(Error::FullFifo);
             }
+
             write_count = std::cmp::min(self.fifo_capacity(), input.len());
-            if write_count > 0 {
-                self.in_buffer.extend(&input[0..write_count]);
-                self.set_lsr_rda_bit();
-                self.received_data_interrupt().map_err(Error::Trigger)?;
-            }
+            self.in_buffer.extend(&input[0..write_count]);
+            self.set_lsr_rda_bit();
+            self.received_data_interrupt().map_err(Error::Trigger)?;
         }
         Ok(write_count)
     }
@@ -630,13 +743,25 @@ mod tests {
         let mut serial = Serial::new(intr_evt.try_clone().unwrap(), sink());
 
         serial.write(IER_OFFSET, IER_RDA_BIT).unwrap();
+
+        serial.enqueue_raw_bytes(&[]).unwrap();
+        // When enqueuing 0 bytes, the serial should neither raise an interrupt,
+        // nor set the `DATA_READY` bit.
+        assert_eq!(
+            intr_evt.read().unwrap_err().kind(),
+            io::ErrorKind::WouldBlock
+        );
+        let mut lsr = serial.read(LSR_OFFSET);
+        assert_eq!(lsr & LSR_DATA_READY_BIT, 0);
+
+        // Enqueue a non-empty slice.
         serial.enqueue_raw_bytes(&RAW_INPUT_BUF).unwrap();
 
         // Verify the serial raised an interrupt.
         assert_eq!(intr_evt.read().unwrap(), 1);
 
         // `DATA_READY` bit should've been set by `enqueue_raw_bytes()`.
-        let mut lsr = serial.read(LSR_OFFSET);
+        lsr = serial.read(LSR_OFFSET);
         assert_ne!(lsr & LSR_DATA_READY_BIT, 0);
 
         // Verify reading the previously pushed buffer.
@@ -817,8 +942,14 @@ mod tests {
         assert_eq!(written_bytes, FIFO_SIZE);
         assert_eq!(serial.in_buffer.len(), FIFO_SIZE);
 
-        // A subsequent call to `enqueue_raw_bytes` fails because the fifo is
-        // now full.
+        // A subsequent call to `enqueue_raw_bytes` with an empty slice should not fail,
+        // even though the fifo is now full.
+        let written_bytes = serial.enqueue_raw_bytes(&[]).unwrap();
+        assert_eq!(written_bytes, 0);
+        assert_eq!(serial.in_buffer.len(), FIFO_SIZE);
+
+        // A subsequent call to `enqueue_raw_bytes` with a non-empty slice fails because
+        // the fifo is now full.
         let one_byte_input = [1u8];
         match serial.enqueue_raw_bytes(&one_byte_input) {
             Err(Error::FullFifo) => (),
@@ -914,4 +1045,69 @@ mod tests {
         // THR empty interrupt was raised nevertheless.
         assert_eq!(iir & IIR_THR_EMPTY_BIT, IIR_THR_EMPTY_BIT);
     }
+
+    #[test]
+    fn test_serial_state_default() {
+        let intr_evt = EventFd::new(libc::EFD_NONBLOCK).unwrap();
+        let serial = Serial::new(intr_evt, Vec::new());
+
+        assert_eq!(serial.state(), SerialState::default());
+    }
+
+    #[test]
+    fn test_from_state_with_too_many_bytes() {
+        let mut state = SerialState::default();
+        let too_many_bytes = vec![1u8; 128];
+
+        state.in_buffer.extend(too_many_bytes);
+
+        let intr_evt = EventFd::new(libc::EFD_NONBLOCK).unwrap();
+        let serial = Serial::from_state(&state, intr_evt, NoEvents, sink());
+
+        assert!(matches!(serial, Err(Error::FullFifo)));
+    }
+
+    #[test]
+    fn test_from_state_with_pending_thre_interrupt() {
+        let intr_evt = EventFd::new(libc::EFD_NONBLOCK).unwrap();
+        let mut serial = Serial::new(intr_evt.try_clone().unwrap(), sink());
+
+        serial.write(IER_OFFSET, IER_THR_EMPTY_BIT).unwrap();
+        serial.write(DATA_OFFSET, b'a').unwrap();
+        assert_eq!(intr_evt.read().unwrap(), 1);
+
+        let state = serial.state();
+        let mut serial_after_restore =
+            Serial::from_state(&state, intr_evt.try_clone().unwrap(), NoEvents, sink()).unwrap();
+
+        let ier = serial_after_restore.read(IER_OFFSET);
+        assert_eq!(ier & IER_UART_VALID_BITS, IER_THR_EMPTY_BIT);
+        let iir = serial_after_restore.read(IIR_OFFSET);
+        assert_ne!(iir & IIR_THR_EMPTY_BIT, 0);
+
+        // Verify the serial raised an interrupt again.
+        assert_eq!(intr_evt.read().unwrap(), 1);
+    }
+
+    #[test]
+    fn test_from_state_with_pending_rda_interrupt() {
+        let intr_evt = EventFd::new(libc::EFD_NONBLOCK).unwrap();
+        let mut serial = Serial::new(intr_evt.try_clone().unwrap(), sink());
+
+        serial.write(IER_OFFSET, IER_RDA_BIT).unwrap();
+        serial.enqueue_raw_bytes(&RAW_INPUT_BUF).unwrap();
+        assert_eq!(intr_evt.read().unwrap(), 1);
+
+        let state = serial.state();
+        let mut serial_after_restore =
+            Serial::from_state(&state, intr_evt.try_clone().unwrap(), NoEvents, sink()).unwrap();
+
+        let ier = serial_after_restore.read(IER_OFFSET);
+        assert_eq!(ier & IER_UART_VALID_BITS, IER_RDA_BIT);
+        let iir = serial_after_restore.read(IIR_OFFSET);
+        assert_ne!(iir & IIR_RDA_BIT, 0);
+
+        // Verify the serial raised an interrupt again.
+        assert_eq!(intr_evt.read().unwrap(), 1);
+    }
 }
